unit DiggerComponent;

interface

{$IF DEFINED(CLR)}
  {$MESSAGE HINT 'Compiling for .NET'}
{$ELSEIF DEFINED(MSWINDOWS)}
  {$MESSAGE HINT 'Compiling for Windows'}
{$ELSEIF DEFINED(LINUX)}
  {$MESSAGE HINT 'Compiling for Linux'}
{$IFEND}

{$IF DEFINED(MSWINDOWS)}
uses Messages, Windows, SysUtils, Classes, Controls, ExtCtrls, Types, Graphics;
{$ELSEIF DEFINED(LINUX)}
uses
  Qt, Types, SysUtils, Classes, QControls, QExtCtrls, QGraphics;
{$MESSAGE WARN 'Linux version not tested!'} 
{$ELSEIF DEFINED(CLR)}
uses System.Drawing, System.Windows.Forms, System.ComponentModel;
{$IFEND}

type

  Directions = (Left, Right, Up, Down, None);

{$IF NOT DEFINED(CLR)}
  Point = TPoint;
  Bitmap = TBitmap;
  Graphics = TCanvas;
  Rectangle = TRect;
{$ELSEIF DEFINED(CLR)}
  TNotifyEvent = procedure (Sender: TObject) of object;
  TTimer = Timer;

  TBrush = SolidBrush;
{$IFEND}

  DiggerFont = class
  protected
    FPosition: Point;
    FSize: Size;
    characters: array of Bitmap;
  public
    constructor Create(ABitmap: Bitmap; ARect: Rectangle);
{$IF NOT DEFINED(CLR)}
    destructor Destroy; override;
{$IFEND}
    procedure Reset;
    procedure Write(AGraphics: Graphics; Text: string);
    procedure NewLine;
  end;

  Sprite = class
  public
    Name: string;
    Invalidate: Boolean;
    constructor Create; virtual;
{$IF NOT DEFINED(CLR)}
    destructor Destroy; override;
{$IFEND}
    function get_ImageIndex: Integer; virtual;
    property ImageIndex: Integer read get_ImageIndex;
  end;

  SpriteGhost = class;
  SpriteDigger = class;

  Position = class
  public
    x, y: Integer;
    constructor Create(Ax, Ay: Integer); overload;
    constructor Create(APosition: Position); overload;
    function Equal(APosition: Position): Boolean;
  end;

  Level = class
  private
    FTime: Integer;
    FOnCompleted: TNotifyEvent;
  protected
    ADigger: SpriteDigger;
    map: array of array of Sprite;
    ghosts: array of SpriteGhost;
    Data: array of Byte;
    FInitTime,
    Number, Diamonds, Collected, Score: Integer;
    procedure SetTime(Value: Integer);
{$IF NOT DEFINED(CLR)}
    procedure ClearSprites;
{$IFEND}
  public
    constructor Create;
{$IF NOT DEFINED(CLR)}
    destructor Destroy; override;
{$IFEND}
    procedure ExitLevel;
    procedure ResetLevel;
    procedure CreateSprite(x, y, id: Integer);
    function GetSprite(x, y: Integer): Sprite;
    function GetSpritePosition(s: Sprite): Position;
    procedure Load(const ATime: Integer);
    procedure SetSprite(x, y: Integer; s: Sprite);
    procedure Tick1;
    procedure Tick4;
  property Time: Integer read FTime write SetTime;
  property OnCompleted: TNotifyEvent read FOnCompleted write FOnCompleted;
  end;

  SpriteDigger = class(Sprite)
  protected
    stone, Keys: array of Boolean;
    direction: Directions;
    step: Integer;
    IsAlive: Boolean;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure Animate(ALevel: Level; x, y: Integer);
    procedure Die;
    function get_ImageIndex: Integer; override;
    procedure Move(ALevel: Level);
  end;

  SpriteGhost = class(Sprite)
  public
    IsAlive: Boolean;
    direction: Directions;
    procedure Blast(ALevel: Level);
    constructor Create; override;
    procedure Die;
    function get_ImageIndex: Integer; override;
    procedure Move(ALevel: Level); virtual; abstract;
  end;

  {$IF DEFINED(CLR)}
  TShortCut = System.Windows.Forms.Shortcut;
  TComponent = System.ComponentModel.Component;
  {$IFEND}

{$IF NOT DEFINED(CLR)}
  THotKeys = class(TComponent)
  private
    FToggleMagnification: TShortCut;
    FGameStart: TShortCut;
    FGameRestart: TShortCut;
    FGamePause: TShortCut;
    FGameReset: TShortCut;
    FGameResume: TShortCut;
  published
    property Start: TShortCut read FGameStart write FGameStart;
    property Resume: TShortCut read FGameResume write FGameResume;
    property Restart: TShortCut read FGameRestart write FGameRestart;
    property Pause: TShortCut read FGamePause write FGamePause;
    property ToggleMagnification: TShortCut read FToggleMagnification write FToggleMagnification;
    property Reset: TShortCut read FGameReset write FGameReset;
  end;
{$IFEND}

{$IF DEFINED(MSWINDOWS)}
  TDiggerPanel = class(TGraphicControl)
{$ELSEIF DEFINED(LINUX)}
  TDiggerPanel = class(TWinControl)
{$ELSEIF DEFINED(CLR)}
  TDiggerPanel = class(System.Windows.Forms.Panel)
  {$REGION 'Designer Managed Code'}
  strict private
    /// <summary>
    /// Required designer variable.
    /// </summary>
    Components: System.ComponentModel.Container;
    /// <summary>
    /// Required method for Designer support - do not modify
    /// the contents of this method with the code editor.
    /// </summary>
  {$ENDREGION}
{$IFEND}
    procedure InitializeComponent;
  private
    procedure SetMagnification(const Value: Integer);
    procedure SetLives(const Value: Integer);
    procedure SetTime(const Value: Integer);
    procedure SetHookKeys(const Value: Boolean);
    procedure UpdateHookKeys;
    procedure UpdateMagnification;
  protected
    { Protected declarations }
    FFont: DiggerFont;
    FSprites: array of Bitmap;
   {$IF NOT DEFINED(CLR)}
      Timer: TTimer;
   {$ELSE}
      Timer: Timer;
   {$IFEND}
    FGameStarted: Boolean;
    keysRelease, keysDelay: array of Boolean;
    FChangingMagnification,
    FHookKeys, FCompletedGame: Boolean;
    FMagnification: Integer;
    FLevel: Level;
    FTime, tick, FInitLives, FLives: Integer;
    FOnMagnificationChanged: TNotifyEvent;

    function MagnifyBitmap(source: Bitmap): Bitmap;
{$IF NOT DEFINED(CLR)}
    procedure FillRectangle(ABrush: TBrush; ARect: Rectangle);
    procedure Timer_Tick(Sender: TObject);
{$ELSE}
    procedure Timer_Tick(sender: &Object; pe: EventArgs);
{$IFEND}

{$IF DEFINED(MSWINDOWS)} // Not tested for Linux compliance
    procedure WMEraseBkgnd(var Message: TMessage); message WM_ERASEBKGND;
    procedure Paint; override;
    procedure Invalidate(Rect: TRect); reintroduce;
    procedure FormKeyDown(Sender: TObject; var Key: Word;
    Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
    Shift: TShiftState);
{$ELSEIF DEFINED(LINUX)}
    procedure PaintRequest; override;
{$ELSEIF DEFINED(CLR)}
  strict protected
    procedure OnPaintBackground(e: PaintEventArgs); override;
    procedure OnPaint(e: PaintEventArgs); override;
    procedure FormKeyUp(Sender: TObject; e: KeyEventArgs);
    procedure FormKeyDown(Sender: TObject; e: KeyEventArgs);
{$IFEND}
    procedure InitializeSettings;
    procedure DoGameCompleted(Sender: TObject);
    procedure Loaded; {$IF NOT DEFINED(CLR)}override;{$IFEND}
  public
    { Public declarations }
{$IF NOT DEFINED(CLR)}
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
{$ELSE}
  strict protected
    procedure Dispose(Disposing: Boolean); override;
    procedure InitLayout; override;
  public
    constructor Create;
{$IFEND}
    procedure Pause;
    procedure Start;
    procedure Restart;
    procedure ResetLives;
    procedure Reset;
{$IF NOT DEFINED(CLR)}
    procedure ProcessKeyUp(Key: Word);
    procedure ProcessKeyDown(Key: Word);
    procedure InitLayout;
{$ELSE}
    procedure ProcessKeyUp(Key: Keys);
    procedure ProcessKeyDown(Key: Keys);
{$IFEND}
  published
    {$IF DEFINED(CLR)}[DefaultValue(2)]{$IFEND}
    property Magnification: Integer read FMagnification write SetMagnification default 2;
    {$IF DEFINED(CLR)}[DefaultValue(600)]{$IFEND}
    property Lives: Integer read FLives write SetLives default 600;
    {$IF DEFINED(CLR)}[DefaultValue(5000)]{$IFEND}
    property Time: Integer read FTime write SetTime default 5000;
    property OnMagnificationChanged: TNotifyEvent read FOnMagnificationChanged write
      FOnMagnificationChanged;
    {$IF DEFINED(CLR)}[DefaultValue(True)]{$IFEND}
    property HookKeys: Boolean read FHookKeys write SetHookKeys default True;
  end;

var
  GKeys: array of Word;

{$IF NOT DEFINED(CLR)}
procedure Register;
{$IFEND}

implementation
{$IF DEFINED(MSWINDOWS)}
  uses Forms; //, MemCheck;
{$ELSEIF DEFINED(LINUX)}
  uses QForms;
{$ELSEIF DEFINED(CLR)}
  uses System.Drawing.Imaging, System.IO, System.Reflection,
    System.Diagnostics, SysUtils;
{$IFEND}

{$IF NOT DEFINED(CLR)}
procedure Register;
begin
  RegisterComponents('Games', [TDiggerPanel]);
end;
{$IFEND}

var
  GMagnification: Integer;
{$IF NOT DEFINED(CLR)}
  GSpriteList: array of Sprite;
{$IFEND}

function MagnifyRectangle(x, y, dx, dy: Integer): Rectangle;
var
  X1, Y1: Integer;
begin
  X1 := x * GMagnification;
  Y1 := y * GMagnification;
{$IF NOT DEFINED(CLR)}
  Result := Rect(X1, Y1, X1 + (dx * GMagnification), Y1 + (dy * GMagnification));
{$ELSE}
  Result := Rectangle.Create(X1, Y1, (dx * GMagnification), (dy * GMagnification));
{$IFEND}
end;

type

{$IF NOT DEFINED(CLR)}
  Color = class
  public
    class function FromArgb(R, G, B: Integer): TColor;
  end;

  SolidBrush = class
  public
    class function Create(AColor: TColor): TBrush;
  end;
{$IFEND}

  SpriteGround = class(Sprite)
  public
    function get_ImageIndex: Integer; override;
  end;

  SpriteNothing = class(Sprite);
  Buffer = class(Sprite);
  SpriteUvStone = class(Sprite);

  SpriteExit = class(Sprite)
  public
    function get_ImageIndex: Integer; override;
  end;

  SpriteChanger = class(Sprite)
  public
    function get_ImageIndex: Integer; override;
  end;

  Gravitation = class(Sprite)
  public
    procedure Move(ALevel: Level; x, y: Integer);
    procedure Simulate(ALevel: Level; x, y: Integer);
  end;

  Marker = class(Gravitation)
  public
    function get_ImageIndex: Integer; override;
  end;

  SpriteStone = class(Gravitation)
  public
    function get_ImageIndex: Integer; override;
  end;

  SpriteDiamond = class(Gravitation)
  public
    class procedure Blink;
    function get_ImageIndex: Integer; override;
  end;

  SpriteGhost180 = class(SpriteGhost)
  public
    procedure Move(ALevel: Level); override;
  end;

  SpriteGhost90L = class(SpriteGhost)
  public
    procedure Move(ALevel: Level); override;
  end;

  SpriteGhost90LR = class(SpriteGhost)
  public
    LastTurn: Directions;
    constructor Create; override;
    procedure Move(ALevel: Level); override;
  end;

  SpriteGhost90R = class(SpriteGhost)
  public
    procedure Move(ALevel: Level); override;
  end;

  SpriteWall = class(Sprite)
  public
    function get_ImageIndex: Integer; override;
  end;

{$IF NOT DEFINED(CLR)}
{ Color }

class function Color.FromArgb(R, G, B: Integer): TColor;
begin
  Result := (B * $10000) + (G * $100) + R;
end;

{ SolidBrush }

class function SolidBrush.Create(AColor: TColor): TBrush;
begin
  Result := TBrush.Create;
  Result.Color := AColor;
end;
{$IFEND}

// Resource embedding
{$IF NOT DEFINED(CLR)}
{$R Level.res}
{$R Font.res}
{$R Sprite.res}
{$ELSE}
{$R 'Level.bin'}
{$R 'Font.png'}
{$R 'Sprite.png'}
{$IFEND}

{ Font }

constructor DiggerFont.Create(ABitmap: Bitmap; ARect: Rectangle);
var
  LCursor: Rectangle;
  AHeight, AWidth, I: Integer;
begin
  inherited Create;
{$IF DEFINED(CLR)}
  FPosition := Point.Create(0, 0);
{$ELSE}
  FPosition := Types.Point(0, 0);
{$IFEND}
  SetLength(characters, 59);
  LCursor := ARect;
  AHeight := ARect.Bottom - ARect.Top;
  AWidth := ARect.Right - ARect.Left;

{$IF DEFINED(CLR)}
  FSize.Width := AWidth;
  FSize.Height := AHeight;
{$ELSE}
  FSize.cx := AWidth;
  FSize.cy := AHeight;
{$IFEND}
{$IF DEFINED(CLR)}
  LCursor := Rectangle.Create(0, 0, FSize.Width, FSize.Height);
{$IFEND}
  for I := Low(characters) to High(characters) do
    begin
{$IF DEFINED(CLR)}
      characters[I] := ABitmap.Clone(LCursor, PixelFormat.Format24bppRgb);
      characters[I].MakeTransparent(Color.FromArgb(0, 0, 0));
      LCursor.Y := LCursor.Y + LCursor.Height;
{$ELSE}
//    characters[I] := ABitmap.Clone(LCursor, PixelFormat.Format24bppRgb);
      characters[I] := TBitmap.Create;
      characters[I].Width := AWidth;
      characters[I].Height := AHeight;
      characters[I].Canvas.CopyRect(Rect(0, 0, AWidth, AHeight),
        ABitmap.Canvas, LCursor);
//    end clone
      characters[I].TransparentColor := TColor($000000);
      characters[I].Transparent := True;
      characters[I].TransparentMode := tmAuto;
      LCursor.Top := LCursor.Top + AHeight;
      LCursor.Bottom := LCursor.Top + AHeight;
{$IFEND}
    end;

  {$IF NOT DEFINED(CLR)}
  ABitmap.Free;
  {$IFEND}
  
end;

{$IF NOT DEFINED(CLR)}
destructor DiggerFont.Destroy;
var
  I: Integer;
begin
  for I := Low(characters) to High(characters) do
    FreeAndNil(characters[I]);
  SetLength(characters, 0);
  inherited;
end;
{$IFEND}

procedure DiggerFont.NewLine;
begin
  FPosition.X := 0;
{$IF DEFINED(CLR)}
  FPosition.Y := FPosition.Y + FSize.Height;
{$ELSE}
  FPosition.Y := FPosition.Y + FSize.cy;
{$IFEND}
end;

procedure DiggerFont.Reset;
begin
{$IF DEFINED(CLR)}
  FPosition := Point.Create(0, 0);
{$ELSE}
  FPosition := Types.Point(0, 0);
{$IFEND}
end;

procedure DiggerFont.Write(AGraphics: Graphics; Text: string);
var
{$IF DEFINED(CLR)}
  redBrush: SolidBrush;
{$IFEND}
  I: Integer;
begin
{$IF DEFINED(CLR)}
  redBrush := SolidBrush.Create(Color.FromArgb($92, $02, $05));
  for I := 1 to Length(Text) do
    begin
      AGraphics.FillRectangle(redBrush, Rectangle.Create(FPosition.X, FPosition.Y, FSize.Width, FSize.Height));
      AGraphics.DrawImageUnscaled(characters[Ord(Text[I]) - 32], FPosition.X, FPosition.Y);
      FPosition.X := FPosition.X + FSize.Width;
    end;
{$ELSE}
  for I := 1 to Length(Text) do
    begin
      AGraphics.Draw(FPosition.X, FPosition.Y, characters[Ord(Text[I]) - 32]);
      FPosition.X := FPosition.X + FSize.cx;
    end;
{$IFEND}
end;

{ TDiggerPanel }

{$IF NOT DEFINED(CLR)}
constructor TDiggerPanel.Create(AOwner: TComponent);
begin
  inherited;
  InitializeComponent;
end;

destructor TDiggerPanel.Destroy;
var
  I: Integer;
begin
  FreeAndNil(Timer);
  FreeAndNil(FFont);
  FreeAndNil(FLevel);
  SetLength(keysRelease, 0);
  SetLength(keysDelay, 0);
  for I := Low(FSprites) to High(FSprites) do
    FreeAndNil(FSprites[I]);
  SetLength(FSprites, 0);
  inherited;
end;

procedure TDiggerPanel.FillRectangle(ABrush: TBrush; ARect: TRect);
begin
  Canvas.Brush := ABrush;
  Canvas.FillRect(ARect);
end;

procedure TDiggerPanel.Invalidate(Rect: TRect);
begin
  OffsetRect(Rect, Left, Top);
{$IF DEFINED(LINUX)}
  InvalidateRect(@Rect, False); // Linux
{$ELSEIF DEFINED(MSWINDOWS)}
  InvalidateRect(Parent.Handle, @Rect, False); // Windows
{$IFEND}
end;
{$ELSE}
constructor TDiggerPanel.Create;
begin
  inherited;
  InitializeComponent;
end;

procedure TDiggerPanel.Dispose(Disposing: Boolean);
begin
  if Disposing then
  begin
    if Components <> nil then
      Components.Dispose();
  end;
  inherited Dispose(Disposing);
end;
{$IFEND}

procedure TDiggerPanel.InitLayout;
begin
  UpdateHookKeys;
  UpdateMagnification;
end;

procedure TDiggerPanel.InitializeComponent;
begin
  Time := 5000;
  Lives := 999;
  Width := 320;
  Height := 256;
  HookKeys := True;
  FChangingMagnification := False;
  Magnification := 1;
  FCompletedGame := False;
  FGameStarted := False;
end;

procedure TDiggerPanel.InitializeSettings;
var
{$IF DEFINED(CLR)}
  Assm: Assembly;
  sprites: Bitmap;
{$ELSE}
  AWidth, AHeight: Integer;
{$IFEND}
  LCursor: Rectangle;
  FontBitmap, SpriteBitmap, MagnifiedSprite: Bitmap;
  I: Integer;
begin
  inherited;

{$IF NOT DEFINED(CLR)}
  FontBitmap := TBitmap.Create;
  FontBitmap.LoadFromResourceID(HInstance, 101);

  if Assigned(FFont) then
    FFont.Free;
{$ELSE}
  // Assm := GetType.Assembly;
  Assm := Assembly.GetExecutingAssembly;
  FontBitmap := Bitmap.Create(Assm.GetManifestResourceStream('Font.png'));
  sprites := MagnifyBitmap(Bitmap.Create(Assm.GetManifestResourceStream('Sprite.png')));
{$IFEND}

  FFont := DiggerFont.Create(MagnifyBitmap(FontBitmap), MagnifyRectangle(0, 0, 8, 8));

  if not Assigned(FSprites) then
    SetLength(FSprites, 34);

//  if not FChangingMagnification then
  begin
  {$IF NOT DEFINED(CLR)}
  if Assigned(FLevel) then
    begin
      FLevel.Free;
      FLevel := nil;
    end;
  {$IFEND}

  FLevel := Level.Create;

  FLevel.OnCompleted := DoGameCompleted;
  end;

  tick := 0;
  SetLength(keysRelease, 4);
  SetLength(keysDelay, 4);

  LCursor := MagnifyRectangle(0, 0, 16, 16);
{$IF NOT DEFINED(CLR)}
// WARNING! DUE TO A BUG IN LoadFromResourcexxx, A NEW TBitmap should be
// created instead of re-using an existing TBitmap
  SpriteBitmap := TBitmap.Create;
  SpriteBitmap.LoadFromResourceID(HInstance, 102);

  AWidth := LCursor.Right - LCursor.Left;
  AHeight := LCursor.Bottom - LCursor.Top;
{$ELSE}
  SpriteBitmap := Bitmap.Create(Assm.GetManifestResourceStream('Sprite.png'));
{$IFEND}
  MagnifiedSprite := MagnifyBitmap(SpriteBitmap);

  for I := Low(FSprites) to High(FSprites) do
    begin
//    sprites.Clone(LCursor, PixelFormat.Format32bppArgb);

      {$IF NOT DEFINED(CLR)}
      if Assigned(FSprites[I]) then
        FreeAndNil(FSprites[I]);

      FSprites[I] := TBitmap.Create;
      FSprites[I].Width := AWidth;
      FSprites[I].Height := AHeight;
      FSprites[I].Canvas.CopyRect(Rect(0, 0, AWidth, AHeight),
        MagnifiedSprite.Canvas, LCursor);

      LCursor.Left := LCursor.Left + AWidth;
      LCursor.Right := LCursor.Left + AWidth;
      {$ELSE}
      FSprites[I] := sprites.Clone(LCursor, PixelFormat.Format32bppArgb);
      LCursor.X := LCursor.X + LCursor.Width;
      {$IFEND}

    end;

//  if not FChangingMagnification then
    begin
      FLevel.Load(FTime);
      ResetLives;
    end;

  {$IF NOT DEFINED(CLR)}
  SpriteBitmap.Free;
  MagnifiedSprite.Free;
  FontBitmap.Free;

//  if not FChangingMagnification then
    begin
       if Assigned(Timer) then
         Timer.Free;
      Timer := TTimer.Create(Self);
      Timer.Enabled := False;
      Timer.Interval := 40;
      Timer.OnTimer := Timer_Tick;
    end;
  {$ELSE}
//  if not FChangingMagnification then
    begin
      Timer := TTimer.Create;
      Timer.Interval := 40;
      Include(Timer.Tick, Timer_Tick);
    end;
  {$IFEND}

end;

procedure TDiggerPanel.DoGameCompleted(Sender: TObject);
begin
  Timer.Enabled := False;
  FCompletedGame := True;
end;

{$IF NOT DEFINED(CLR)}
procedure TDiggerPanel.FormKeyDown(Sender: TObject; var Key: Word;
    Shift: TShiftState);
begin
  ProcessKeyDown(Key);
end;

procedure TDiggerPanel.FormKeyUp(Sender: TObject; var Key: Word;
    Shift: TShiftState);
begin
  SetLength(GKeys, Length(GKeys)+1);
  GKeys[High(GKeys)] := Key;
  ProcessKeyUp(Key);
end;
{$ELSE}
procedure TDiggerPanel.FormKeyDown(Sender: TObject; e: KeyEventArgs);
begin
  ProcessKeyDown(e.KeyCode);
end;

procedure TDiggerPanel.FormKeyUp(Sender: TObject; e: KeyEventArgs);
begin
  ProcessKeyUp(e.KeyCode);
end;
{$IFEND}

{$IF DEFINED(LINUX)}
const
  VK_LEFT   = KEY_LEFT;
  VK_RIGHT  = KEY_RIGHT;
  VK_UP     = KEY_UP;
  VK_DOWN   = KEY_DOWN;
  VK_ESCAPE = KEY_ESCAPE;
  VK_HOME   = KEY_HOME;
{$IFEND}

{$IF NOT DEFINED(CLR)}
procedure TDiggerPanel.ProcessKeyDown(Key: Word);
{$ELSE}
procedure TDiggerPanel.ProcessKeyDown(Key: Keys);
{$IFEND}
begin
  case Key of
    {$IF NOT DEFINED(CLR)}VK_LEFT:{$ELSE}Keys.Left:{$IFEND}
      begin
        FLevel.ADigger.Keys[0] := True;
        keysDelay[0] := False;
      end;
    {$IF NOT DEFINED(CLR)}VK_RIGHT:{$ELSE}Keys.Right:{$IFEND}
      begin
        FLevel.ADigger.Keys[1] := True;
        keysDelay[1] := False;
      end;
    {$IF NOT DEFINED(CLR)}VK_UP:{$ELSE}Keys.Up:{$IFEND}
      begin
        FLevel.ADigger.Keys[2] := True;
        keysDelay[2] := False;
      end;
    {$IF NOT DEFINED(CLR)}VK_DOWN:{$ELSE}Keys.Down:{$IFEND}
      begin
        FLevel.ADigger.Keys[3] := True;
        keysDelay[3] := False;
      end;
    {$IF NOT DEFINED(CLR)}VK_ESCAPE:{$ELSE}Keys.Escape:{$IFEND}
       if FGameStarted then Restart;
    {$IF NOT DEFINED(CLR)}VK_HOME:{$ELSE}Keys.Home:{$IFEND}
      if FLevel.Number < 29 then
        begin
          Inc(FLevel.Number);
          FLevel.Load(FTime);
          inherited Invalidate;
        end;
  end;
end;

{$IF NOT DEFINED(CLR)}
procedure TDiggerPanel.ProcessKeyUp(Key: Word);
{$ELSE}
procedure TDiggerPanel.ProcessKeyUp(Key: Keys);
{$IFEND}
begin
  case Key of
    {$IF NOT DEFINED(CLR)}VK_LEFT:{$ELSE}Keys.Left:{$IFEND}
      if keysDelay[0] then
        FLevel.ADigger.Keys[0] := False else
        keysRelease[0] := True;
    {$IF NOT DEFINED(CLR)}VK_RIGHT:{$ELSE}Keys.Right:{$IFEND}
      if keysDelay[1] then
        FLevel.ADigger.Keys[1] := False else
        keysRelease[1] := True;
    {$IF NOT DEFINED(CLR)}VK_UP:{$ELSE}Keys.Up:{$IFEND}
      if keysDelay[2] then
        FLevel.ADigger.Keys[2] := False else
        keysRelease[2] := True;
    {$IF NOT DEFINED(CLR)}VK_DOWN:{$ELSE}Keys.Down:{$IFEND}
      if keysDelay[3] then
        FLevel.ADigger.Keys[3] := False else
        keysRelease[3] := True;
  end;
end;

procedure TDiggerPanel.Loaded;
begin
  // InitializeSettings;
  InitLayout;
end;

function TDiggerPanel.MagnifyBitmap(source: Bitmap): Bitmap;
var
  x, y, ix, iy: Integer;
begin
{$IF NOT DEFINED(CLR)}
  Result := Bitmap.Create;
  Result.Width := source.Width * Magnification;
  Result.Height := source.Height * Magnification;
{$ELSE}
  Result := Bitmap.Create(source.Width * Magnification, source.Height * Magnification);
{$IFEND}
  for x := 0 to source.Width - 1 do
    for y := 0 to source.Height - 1 do
      for ix := 0 to magnification - 1 do
        for iy := 0 to magnification - 1 do
{$IF NOT DEFINED(CLR)}
          Result.Canvas.Pixels[x * magnification + ix, y * magnification + iy] := source.Canvas.Pixels[x, y];
{$ELSE}
          Result.SetPixel(x * magnification + ix, y * magnification + iy, source.GetPixel(x, y));
{$IFEND}

end;

{$IF DEFINED(CLR)}
procedure TDiggerPanel.OnPaintBackground(e: PaintEventArgs);
begin
 // Empty
end;

procedure TDiggerPanel.OnPaint(e: PaintEventArgs);
var
  LRedBrush, LBlueBrush, LCyanBrush: SolidBrush;
  LRectangle: Rectangle;
  Canvas: Graphics;
{$ELSE}
procedure TDiggerPanel.Paint;
var
  LRedBrush, LBlueBrush, LCyanBrush: TBrush;
  LRectangle: TRect;
{$IFEND}
var
  y, x: Integer;
  I: Integer;
begin

{$IF DEFINED(CLR)}
  Canvas := e.Graphics;

  LBlueBrush := SolidBrush.Create(Color.FromArgb($04, $02, $8F));
  LCyanBrush := SolidBrush.Create(Color.FromArgb($2D, $E7, $C0));
  LRedBrush := SolidBrush.Create(Color.FromArgb($92, $02, $05));


  Canvas.FillRectangle(LBlueBrush, MagnifyRectangle(0, 0, 320, 8));
  Canvas.FillRectangle(LCyanBrush, MagnifyRectangle(0, 2, 320, 4));


  //Canvas.FillRectangle(LRedBrush, MagnifyRectangle(0, 8, 320, 16));


  Canvas.FillRectangle(LBlueBrush, MagnifyRectangle(0, 24, 320, 8));
  Canvas.FillRectangle(LCyanBrush, MagnifyRectangle(0, 26, 320, 4));

{$ELSE}
  LBlueBrush := SolidBrush.Create(DiggerComponent.Color.FromArgb($04, $02, $8F));
  LCyanBrush := SolidBrush.Create(DiggerComponent.Color.FromArgb($2D, $E7, $C0));
  LRedBrush := SolidBrush.Create(DiggerComponent.Color.FromArgb($92, $02, $05));

  FillRectangle(LBlueBrush, MagnifyRectangle(0, 0, 320, 8));
  FillRectangle(LBlueBrush, MagnifyRectangle(0, 24, 320, 8));

  FillRectangle(LCyanBrush, MagnifyRectangle(0, 2, 320, 4));
  FillRectangle(LCyanBrush, MagnifyRectangle(0, 26, 320, 4));

  FillRectangle(LRedBrush, MagnifyRectangle(0, 8, 320, 16));
{$IFEND}

  FFont.Reset;
  FFont.Newline;
  FFont.Write(Canvas, Format('  CAVE:  %2.2d TIME:  %5.5d DIAMONDS:  %2.2d  ', [FLevel.Number + 1, FLevel.Time, FLevel.Diamonds]));
  FFont.Newline;
  FFont.Write(Canvas, Format('  LIVES:%3d SCORE: %5.5d COLLECTED: %2.2d  ', [lives, FLevel.Score, FLevel.Collected]));
  FFont.Newline;

  if not Assigned(FLevel) then
    raise Exception.Create('FLevel not assigned!');

  for y := 0 to 13 do
    begin
      for x := 0 to 19 do
        begin
          LRectangle := MagnifyRectangle(x * 16, y * 16 + 32, 16, 16);
          I := FLevel.GetSprite(x, y).ImageIndex;
{$IF NOT DEFINED(CLR)}
          Canvas.Draw(LRectangle.Left, LRectangle.Top, FSprites[I]);
{$ELSE}
          if (e.ClipRectangle.IntersectsWith(LRectangle)) then
            Canvas.DrawImageUnscaled(FSprites[I], LRectangle.X, LRectangle.Y);
{$IFEND}
        end;
    end;

  if FCompletedGame then
    begin
{$IF NOT DEFINED(CLR)}
      FillRectangle(LRedBrush, MagnifyRectangle(0, 32, 320, 8));
{$ELSE}
      Canvas.FillRectangle(LRedBrush, MagnifyRectangle(0, 32, 320, 8));
{$IFEND}
      FFont.NewLine;
      FFont.Write(Canvas, '            GAME COMPLETED              ');
{$IF NOT DEFINED(CLR)}
      FillRectangle(LBlueBrush, MagnifyRectangle(0, 40, 320, 8));
      FillRectangle(LCyanBrush, MagnifyRectangle(0, 42, 320, 4));
{$ELSE}
      Canvas.FillRectangle(LBlueBrush, MagnifyRectangle(0, 40, 320, 8));
      Canvas.FillRectangle(LCyanBrush, MagnifyRectangle(0, 42, 320, 4));
{$IFEND}
    end;

  {$IF NOT DEFINED(CLR)}
  LCyanBrush.Free;
  LBlueBrush.Free;
  LRedBrush.Free;
  {$IFEND}
end;

{ Sprite }

constructor Sprite.Create;
begin
  inherited;
  Name := ClassName;
{$IF NOT DEFINED(CLR)}
  SetLength(GSpriteList, Length(GSpriteList)+1);
  GSpriteList[High(GSpriteList)] := Self;
{$IFEND}
  Invalidate := True;
end;

{$IF NOT DEFINED(CLR)}
destructor Sprite.Destroy;
var
  I: Integer;
begin
  OutputDebugString(PChar(Name));
  for I := Low(GSpriteList) to High(GSpriteList) do
    if GSpriteList[I] = Self then
      GSpriteList[I] := nil;
  inherited;
end;
{$IFEND}

function Sprite.get_ImageIndex: Integer;
begin
  Result := 0;
end;

{ Position }

constructor Position.Create(Ax, Ay: Integer);
begin
  inherited Create;
  x := Ax;
  y := Ay;
end;

constructor Position.Create(APosition: Position);
begin
  inherited Create;
  x := APosition.x;
  y := APosition.y;
end;

function Position.Equal(APosition: Position): Boolean;
begin
  Result := (x = APosition.x) and (y = APosition.y);
end;

{ Level }
{$IF NOT DEFINED(CLR)}
procedure Level.ClearSprites;
var
  I: Integer;
  Temp: Sprite;
begin
  for I := Low(GSpriteList) to High(GSpriteList) do
    begin
      Temp := GSpriteList[I];
      try
        FreeAndNil(GSpriteList[I]);
      except
        OutputDebugString(PChar(Temp.Name));
      end;
    end;
  SetLength(GSpriteList, 0);
  ADigger := nil;
end;
{$IFEND}

constructor Level.Create;
var
{$IF DEFINED(CLR)}
  AStream: Stream;
  Assm: Assembly;
{$ELSE}
  AStream: TStream;
{$IFEND}
begin
  inherited;
  SetLength(map, 20, 14);
  SetLength(ghosts, 16);

{$IF DEFINED(CLR)}
  Assm := GetType.Assembly;
  AStream := Assm.GetManifestResourceStream('Level.bin');
  SetLength(data, AStream.Length);
  AStream.Read(data, 0, AStream.Length);
{$ELSE}
  AStream := TResourceStream.Create(HInstance, 'Level', RT_RCDATA);

  SetLength(Data, AStream.Size);

  AStream.Read(Data[0], Length(Data));

  AStream.Free;
{$IFEND}
  ResetLevel;
end;

procedure Level.CreateSprite(x, y, id: Integer);
type
  SpriteClass = class of Sprite;
const
  t: array[0..15] of SpriteClass = (
    SpriteNothing, SpriteStone, SpriteGround, SpriteGhost180,
    nil, SpriteDiamond, SpriteWall, SpriteGhost90L,
    nil, SpriteUvStone, SpriteDigger, SpriteGhost90LR,
    SpriteExit, nil, SpriteChanger, SpriteGhost90R
    );
var
  ASprite: Sprite;
begin
  ASprite := t[id].Create;
  SetSprite(x, y, ASprite);
end;

{$IF NOT DEFINED(CLR)}
destructor Level.Destroy;
begin
  ClearSprites;
  SetLength(Data, 0);
  SetLength(ghosts, 0);
  SetLength(map, 0);
  inherited;
end;
{$IFEND}

procedure Level.ExitLevel;
begin
  if Collected < Diamonds then Exit;
  if Number >= 29 then
    begin
      if Assigned(FOnCompleted) then
        FOnCompleted(Self);
    end else
    begin
      Inc(Number);
      Load(FInitTime);
    end;
end;

function Level.GetSprite(x, y: Integer): Sprite;
begin
  if not Assigned(map[x, y]) then
    raise Exception.Create('map empty!');
  Result := map[x, y];
end;

function Level.GetSpritePosition(s: Sprite): Position;
var
  x, y: Integer;
begin
  Result := nil;
  for x := 0 to 19 do
    for y := 0 to 13 do
      if GetSprite(x, y) = s then
        begin
          Result := Position.Create(x, y);
          exit;
        end;
  if Result = nil then
    Result := Position.Create;
end;

procedure Level.Load(const ATime: Integer);
type
  TDirectionsArray = array of Directions;
const
  d: array[0..3] of Directions = (Down, Up, Right, Left);
var
  LLevel: array of Byte;
  AGhost: SpriteGhost;
  I, x, y, g, Ghost, GhostInfo: Integer;
begin

  SetLength(GKeys, 0);

  SetLength(LLevel, 156);
  for I := Low(LLevel) to High(LLevel) do
    LLevel[I] := data[Number * 156 + I];
{$IF NOT DEFINED(CLR)}
  ClearSprites;
{$IFEND}
  for y := 0 to 13 do
    for x := 0 to 9 do
      begin
        CreateSprite(x * 2, y, LLevel[y * 10 + x] shr 4);
        CreateSprite(x * 2 + 1, y, LLevel[y * 10 + x] and $F);
      end;
  x := LLevel[145];
  y := LLevel[146] - 2;
  ADigger := map[x, y] as SpriteDigger;
  for g := Low(ghosts) to High(ghosts) do
    ghosts[g] := nil;
  Ghost := 0;
  for y := 0 to 13 do
    for x := 0 to 19 do
      begin
        if GetSprite(x, y) is SpriteGhost then
          begin
            AGhost := GetSprite(x, y) as SpriteGhost;
            ghosts[Ghost] := AGhost;
            GhostInfo := LLevel[$94 + (Ghost shr 1)];
            if (Ghost and 1) <> 0 then
              begin
                GhostInfo := GhostInfo and $F;
                if AGhost is SpriteGhost90LR then
                  SpriteGhost90LR(AGhost).LastTurn := Right;
              end else
              begin
                GhostInfo := GhostInfo shr 4;
                if AGhost is SpriteGhost90LR then
                  (AGhost as SpriteGhost90LR).LastTurn := Left;
              end;
            AGhost.Direction := d[GhostInfo];
            Inc(Ghost);
          end;
      end;
  Diamonds := (LLevel[147] div $10) * 10 + (LLevel[147] mod $10);
  Collected := 0;
  Time := ATime;
end;

procedure Level.ResetLevel;
begin
  Number := 0;
end;

procedure Level.SetSprite(x, y: Integer; s: Sprite);
begin
  map[x, y] := s;
  s.Invalidate := True;
end;

procedure Level.SetTime(Value: Integer);
begin
  FInitTime := Value;
  FTime := Value;
end;

procedure Level.Tick1;
var
  x, y: Integer;
  s: Sprite;
begin
  SpriteDiamond.Blink;
  for y := 0 to 13 do
    for x := 0 to 19 do
      begin
        s := GetSprite(x, y);
        if s is SpriteDiamond then
          s.Invalidate := True;
      end;
end;

procedure Level.Tick4;
var
  x, y, g: Integer;
begin
  for y := 13 downto 0 do
    for x := 19 downto 0 do
      begin
        if map[x, y] is Buffer then
          SetSprite(x, y, SpriteNothing.Create);
      end;
  if not Assigned(ADigger) then
    begin
      raise Exception.Create('Digger not assigned!');
    end;
  ADigger.Move(Self);
  for y := 13 downto 0 do
    for x := 19 downto 0 do
      begin
        if (map[x, y] is SpriteStone) or
          (map[x, y] is SpriteDiamond) or
          (map[x, y] is SpriteUvStone) then
          Gravitation(map[x, y]).Simulate(Self, x, y);
      end;
  for y := 13 downto 0 do
    for x := 19 downto 0 do
      if (map[x, y] is SpriteStone) or
        (map[x, y] is SpriteDiamond) or
        (map[x, y] is SpriteUvStone) then
        begin
          Gravitation(map[x, y]).Move(Self, x, y);
        end;
  for g := Low(ghosts) to High(ghosts) do
    if Assigned(ghosts[g]) then
      ghosts[g].Move(Self);
  if FTime > 0 then
    Dec(FTime);
  if FTime = 0 then
    ADigger.Die;
end;

{ Digger }

procedure SpriteDigger.Animate(ALevel: Level; x, y: Integer);
begin
  if IsAlive then
    begin
      case direction of
        left:
          begin
            Inc(step);
            if step >= 6 then step := 0;
            ALevel.ADigger.Invalidate := True;
          end;
        Right:
          begin
            Inc(step);
            if step >= 6 then step := 0;
            ALevel.ADigger.Invalidate := True;
          end;
        Up:
          begin
            Inc(step);
            if step >= 2 then step := 0;
            ALevel.ADigger.Invalidate := True;
          end;
        Down:
          begin
            Inc(step);
            if (step >= 2) then step := 0;
            ALevel.ADigger.Invalidate := True;
          end;
        None:
          begin
            Inc(step);
            if (step >= 30) then step := 0;
            ALevel.ADigger.Invalidate := True;
          end;
      end;
    end;
end;

constructor SpriteDigger.Create;
begin
  inherited;
  SetLength(Keys, 4);
  direction := None;
  step := 0;
  IsAlive := True;
  SetLength(stone, 2);
end;

destructor SpriteDigger.Destroy;
begin
  SetLength(stone, 0);
  SetLength(Keys, 0);
  inherited;
end;

procedure SpriteDigger.Die;
begin
  IsAlive := False;
  Invalidate := True;
end;

function SpriteDigger.get_ImageIndex: Integer;
const
  stepleft: array[0..5] of Integer = (16, 17, 18, 19, 18, 17);
  stepright: array[0..5] of Integer = (20, 21, 22, 23, 22, 21);
  stepup: array[0..1] of Integer = (24, 25);
  stepdown: array[0..1] of Integer = (26, 27);
  stepcenter: array[0..29] of Integer = (15, 15, 15, 15, 15, 15, 15, 15, 28, 28, 15, 15, 28, 28, 15, 15, 15, 15, 15, 15, 29, 29, 30, 30, 29, 29, 15, 15, 15, 15);
begin
  if IsAlive then
    begin
      Result := stepcenter[step];
      case direction of
        Left: if step <= High(stepleft) then Result := stepleft[step];
        Right: if step <= High(stepright) then Result := stepright[step];
        Up: if step <= High(stepup) then Result := stepup[step];
        Down: if step <= High(stepdown) then Result := stepdown[step];
      end;
    end else Result := 31;
end;

procedure SpriteDigger.Move(ALevel: Level);
var
  p, d, z: Position;
begin
  {$IF NOT DEFINED(CLR)}
  p := nil; d := nil; z := nil;
  {$IFEND}

  if IsAlive then
    begin
      p := ALevel.GetSpritePosition(Self);
      d := Position.Create(p);
      z := Position.Create(d);
      direction := None;
      if (Keys[0]) then
        begin
          Dec(z.x);
          direction := Left;
        end else
        begin
          stone[0] := False;
          if Keys[1] then
            begin
              Inc(z.x);
              direction := Right;
            end else
            begin
              stone[1] := False;
              if (Keys[2]) then
                begin
                  Dec(z.y);
                  direction := Up;
                end else
                begin
                  if (Keys[3]) then
                    begin
                      Inc(z.y);
                      direction := Down;
                    end;
                end;
            end;
        end;
      if not d.Equal(z) then
        begin
          // Nothing
          if ALevel.GetSprite(z.x, z.y) is SpriteNothing then
            ALevel.SetSprite(d.x, d.y, Self);
          // Diamond
          if ALevel.GetSprite(z.x, z.y) is SpriteDiamond then
            begin
              Inc(ALevel.Collected);
              Inc(ALevel.Score, 3);
            end;
          // Stone
          if ALevel.GetSprite(z.x, z.y) is SpriteStone then
            begin
              if (z.x > d.x) and (ALevel.GetSprite(z.x + 1, z.y) is SpriteNothing) then
                begin
                  if stone[1] then
                    begin
                      ALevel.SetSprite(d.x + 2, d.y, ALevel.GetSprite(d.x + 1, d.y));
                      ALevel.SetSprite(d.x + 1, d.y, SpriteNothing.Create);
                    end;
                  stone[1] := not stone[1];
                end;
              if (z.x < d.x) and (ALevel.GetSprite(z.x - 1, z.y) is SpriteNothing) then
                begin
                  if stone[0] then
                    begin
                      ALevel.SetSprite(d.x - 2, d.y, ALevel.GetSprite(d.x - 1, d.y));
                      ALevel.SetSprite(d.x - 1, d.y, SpriteNothing.Create);
                    end;
                  stone[0] := not stone[0];
                end;
            end;
          // Ground
          if (ALevel.GetSprite(z.x, z.y) is SpriteNothing) or
            (ALevel.GetSprite(z.x, z.y) is SpriteGround) or
            (ALevel.GetSprite(z.x, z.y) is SpriteDiamond) then
            begin
              ALevel.SetSprite(z.x, z.y, Self);
              ALevel.SetSprite(d.x, d.y, Buffer.Create);
            end;
          // Exit
          if ALevel.GetSprite(z.x, z.y) is SpriteExit then
            ALevel.ExitLevel;
          // Ghost
          if ALevel.GetSprite(z.x, z.y) is SpriteGhost then
            ALevel.ADigger.Die;
        end;
      Animate(ALevel, z.x, z.y);
    end;

  {$IF NOT DEFINED(CLR)}
  FreeAndNil(p);
  FreeAndNil(d);
  FreeAndNil(z);
  {$IFEND}
end;

{ Ghost }

procedure SpriteGhost.Blast(ALevel: Level);
var
  LPosition: Position;
  y, x: Integer;
begin
  if IsAlive then
    begin
      LPosition := ALevel.GetSpritePosition(Self);
      for y := LPosition.y - 1 to LPosition.y + 1 do
        for x := LPosition.x - 1 to LPosition.x + 1 do
          begin
            if (x > 0) and (x < 19) and (y > 0) and (y < 13) then
              begin
                if ALevel.GetSprite(x, y) is SpriteDigger then
                  begin
                    ALevel.ADigger.Die;
                  end else
                  begin
                    if ALevel.GetSprite(x, y) is SpriteGhost then
                      begin
                        (ALevel.GetSprite(x, y) as SpriteGhost).Die;
                        Inc(ALevel.Score, 99);
                      end;
                    ALevel.SetSprite(x, y, SpriteNothing.Create);
                  end;
              end;
          end;
      Die;
      {$IF NOT DEFINED(CLR)}
      LPosition.Free;
      {$IFEND}
    end;

end;

constructor SpriteGhost.Create;
begin
  inherited;
  IsAlive := True;
  Direction := None;
end;

procedure SpriteGhost.Die;
begin
  IsAlive := False;
end;

function SpriteGhost.get_ImageIndex: Integer;
begin
  case Direction of
    Left: Result := 4;
    Right: Result := 5;
    Up: Result := 6;
    Down: Result := 3;
  else
    Result := 3;
  end;
end;

{ Ground }

function SpriteGround.get_ImageIndex: Integer;
begin
  Result := 2;
end;

{ Exit }

function SpriteExit.get_ImageIndex: Integer;
begin
  Result := 32;
end;

{ Changer }

function SpriteChanger.get_ImageIndex: Integer;
begin
  Result := 33;
end;

{ Gravitation }

procedure Gravitation.Move(ALevel: Level; x, y: Integer);
var
  dx, dy: Integer;
begin
  dx := x;
  dy := y;
  // Follow gravitation
  if ALevel.GetSprite(x, y + 1) is Marker then
    begin
      dy := y + 1;
    end else
    begin
      // Falls to the left or to the right
      if (ALevel.GetSprite(x, y + 1) is SpriteStone) or
        (ALevel.GetSprite(x, y + 1) is SpriteDiamond) or
        (ALevel.GetSprite(x, y + 1) is SpriteNothing) then
        begin
          if (ALevel.GetSprite(x - 1, y + 1) is Marker) and
            ((ALevel.GetSprite(x - 1, y) is SpriteNothing) or
            (ALevel.GetSprite(x - 1, y) is Marker)) then
            begin
              dx := x - 1;
              dy := y + 1;
            end else
            begin
              if (ALevel.GetSprite(x + 1, y + 1) is Marker) and
                ((ALevel.GetSprite(x + 1, y) is SpriteNothing) or
                (ALevel.GetSprite(x + 1, y) is Marker)) then
                begin
                  dx := x + 1;
                  dy := y + 1;
                end;
            end;
        end;
      // Changer
      if (ALevel.GetSprite(x, y + 1) is SpriteChanger) and
        (ALevel.GetSprite(x, y) is SpriteStone) and
        (ALevel.GetSprite(x, y + 2) is Marker) then
        dy := y + 2;
    end;
  // Move
  if (dx <> x) or (dy <> y) then
    begin
      if (dy - y) = 2 then
        begin
          // changer
          ALevel.SetSprite(dx, dy, SpriteDiamond.Create);
        end else
        begin
          ALevel.SetSprite(dx, dy, ALevel.GetSprite(x, y));
          if ALevel.GetSprite(dx, dy) is SpriteUvStone then
            ALevel.SetSprite(dx, dy, SpriteStone.Create);
        end;
      ALevel.SetSprite(x, y, SpriteNothing.Create);
      // Kill digger if necessary
      if ALevel.GetSprite(dx, dy + 1) is SpriteDigger then
        ALevel.ADigger.Die;
      if ALevel.GetSprite(dx, dy + 1) is SpriteGhost then
        begin
          (ALevel.GetSprite(dx, dy + 1) as SpriteGhost).Blast(ALevel);
        end;
    end;
end;

procedure Gravitation.Simulate(ALevel: Level; x, y: Integer);
var
  dx, dy: Integer;
begin
  dx := x;
  dy := y;
  // Nothing
  if ALevel.GetSprite(x, y + 1) is SpriteNothing then
    begin
      dy := y + 1;
    end else
    begin
      // Stone or diamond
      if (ALevel.GetSprite(x, y + 1) is SpriteStone) or
        (ALevel.GetSprite(x, y + 1) is SpriteDiamond) then
        begin
          if (ALevel.GetSprite(x - 1, y + 1) is SpriteNothing) and (ALevel.GetSprite(x - 1, y) is SpriteNothing) then
            begin
              dx := x - 1;
              dy := y + 1;
            end else
            begin
              if (ALevel.GetSprite(x + 1, y + 1) is SpriteNothing) and (ALevel.GetSprite(x + 1, y) is SpriteNothing) then
                begin
                  dx := x + 1;
                  dy := y + 1;
                end;
            end;
        end;
      // Changer
      if (ALevel.GetSprite(x, y + 1) is SpriteChanger) and (ALevel.GetSprite(x, y) is SpriteStone) and (ALevel.GetSprite(x, y + 2) is SpriteNothing) then
        dy := y + 2;
    end;
  // Mark
  if (dx <> x) or (dy <> y) then
    ALevel.SetSprite(dx, dy, Marker.Create);
end;

{ Marker }

function Marker.get_ImageIndex: Integer;
begin
  Result := 31;
end;

{ Stone }

function SpriteStone.get_ImageIndex: Integer;
begin
  Result := 1;
end;

{ Diamond }
var
  FBlinkCounter: Integer;

class procedure SpriteDiamond.Blink;
begin
  FBlinkCounter := (FBlinkCounter + 1) mod 6;
end;

function SpriteDiamond.get_ImageIndex: Integer;
begin
  Result := (13 - ((FBlinkCounter + 4 * 1) mod 6));
end;

{ Ghost180 }

procedure SpriteGhost180.Move(ALevel: Level);
type
  TPositionArray = array of Position;
var
  w: TPositionArray;
  d, p: Position;
  I: Integer;
begin

{$IF NOT DEFINED(CLR)}
  d := nil; p := nil;
{$IFEND}

  if IsAlive then
    begin
      p := ALevel.GetSpritePosition(Self);
      SetLength(w, 4);
      for I := Low(w) to High(w) do
        w[I] := Position.Create(p);
      case Direction of
        Left:
          begin
            Dec(w[0].x);
            Inc(w[1].x);
          end;
        Right:
          begin
            Inc(w[0].x);
            Dec(w[1].x);
          end;
        Up:
          begin
            Dec(w[0].y);
            Inc(w[1].y);
          end;
        Down:
          begin
            Inc(w[0].y);
            Dec(w[1].y);
          end;
      end;
      for I := 0 to 3 do
        begin
          if not p.Equal(w[i]) then
            begin
              {$IF NOT DEFINED(CLR)}FreeAndNil(d);{$IFEND}
              d := Position.Create(w[i]);
              // Digger
              if (ALevel.GetSprite(d.x, d.y) is SpriteDigger) then
                ALevel.ADigger.Die;
              // Nothing
              if (ALevel.GetSprite(d.x, d.y) is SpriteNothing) then
                begin
                  if (d.x < p.x) then Direction := Left;
                  if (d.x > p.x) then Direction := Right;
                  if (d.y < p.y) then Direction := Up;
                  if (d.y > p.y) then Direction := Down;
                  ALevel.SetSprite(d.x, d.y, Self);
                  ALevel.SetSprite(p.x, p.y, SpriteNothing.Create);
                  Break;
                end;
            end;
        end;

      {$IF NOT DEFINED(CLR)}
      for I := Low(w) to High(w) do
        FreeAndNil(w[I]);
      SetLength(w, 0);

      FreeAndNil(d);
      FreeAndNil(p);

      {$IFEND}

    end;

end;

{ Ghost90L }

procedure SpriteGhost90L.Move(ALevel: Level);
type
  TPositionArray = array of Position;
var
  w: TPositionArray;
  d, p: Position;
  I: Integer;
begin

{$IF NOT DEFINED(CLR)}
  d := nil; p := nil;
{$IFEND}

  if IsAlive then
    begin
      p := ALevel.GetSpritePosition(Self);
      SetLength(w, 4);
      for I := Low(w) to High(w) do
        w[I] := Position.Create(p);
      case Direction of
        Left:
          begin
            Dec(w[0].x);
            Inc(w[1].y);
            Dec(w[2].y);
            Inc(w[3].x);
          end;
        Right:
          begin
            Inc(w[0].x);
            Dec(w[1].y);
            Inc(w[2].y);
            Dec(w[3].x);
          end;
        Up:
          begin
            Dec(w[0].y);
            Dec(w[1].x);
            Inc(w[2].x);
            Inc(w[3].y);
          end;
        Down:
          begin
            Inc(w[0].y);
            Inc(w[1].x);
            Dec(w[2].x);
            Dec(w[3].y);
          end;
      end;
      for I := 0 to 3 do
        begin
          if not p.Equal(w[i]) then
            begin
              {$IF NOT DEFINED(CLR)}FreeAndNil(d);{$IFEND}
              d := Position.Create(w[i]);
          // Digger
              if ALevel.GetSprite(d.x, d.y) is SpriteDigger then
                ALevel.ADigger.Die;
            // Nothing
              if ALevel.GetSprite(d.x, d.y) is SpriteNothing then
                begin
                  if (d.x < p.x) then Direction := Left;
                  if (d.x > p.x) then Direction := Right;
                  if (d.y < p.y) then Direction := Up;
                  if (d.y > p.y) then Direction := Down;
                  ALevel.SetSprite(d.x, d.y, Self);
                  ALevel.SetSprite(p.x, p.y, SpriteNothing.Create);
                  Break;
                end;
            end;
        end;

      {$IF NOT DEFINED(CLR)}
      for I := Low(w) to High(w) do
        FreeAndNil(w[I]);
      SetLength(w, 0);

      FreeAndNil(d);
      FreeAndNil(p);
      {$IFEND}

    end;

end;

{ Ghost90LR }

constructor SpriteGhost90LR.Create;
begin
  inherited;
  LastTurn := None;
end;

procedure SpriteGhost90LR.Move(ALevel: Level);
type
  TPositionArray = array of Position;
var
  w: TPositionArray;
  d, p: Position;
  I: Integer;
  LastDirection: Directions;
begin

{$IF NOT DEFINED(CLR)}
  d := nil; p := nil;
{$IFEND}

  if IsAlive then
    begin
      p := ALevel.GetSpritePosition(Self);
      SetLength(w, 4);
      for I := Low(w) to High(w) do
        w[I] := Position.Create(p);
      case Direction of
        Left:
          begin
            Dec(w[0].x);
            Inc(w[3].x);
            if LastTurn = Left then
              begin
                Dec(w[1].y);
                Inc(w[2].y);
              end else
              begin
                Inc(w[1].y);
                Dec(w[2].y);
              end;
          end;
        Right:
          begin
            Inc(w[0].x);
            Dec(w[3].x);
            if (LastTurn = Left) then
              begin
                Inc(w[1].y);
                Dec(w[2].y);
              end else
              begin
                Dec(w[1].y);
                Inc(w[2].y);
              end;
          end;
        Up:
          begin
            Dec(w[0].y);
            Inc(w[3].y);
            if (LastTurn = Left) then
              begin
                Inc(w[1].x);
                Dec(w[2].x);
              end else
              begin
                Dec(w[1].x);
                Inc(w[2].x);
              end;
          end;
        Down:
          begin
            Inc(w[0].y);
            Dec(w[3].y);
            if (LastTurn = Left) then
              begin
                Dec(w[1].x);
                Inc(w[2].x);
              end else
              begin
                Inc(w[1].x);
                Dec(w[2].x);
              end;
          end;
      end;
      for I := 0 to 3 do
        begin
          if not p.Equal(w[i]) then
            begin
              {$IF NOT DEFINED(CLR)}FreeAndNil(d);{$IFEND}
              d := Position.Create(w[i]);
              // Digger
              if (ALevel.GetSprite(d.x, d.y) is SpriteDigger) then
                ALevel.ADigger.Die;
              // Nothing
              if (ALevel.GetSprite(d.x, d.y) is SpriteNothing) then
                begin
                  LastDirection := Direction;
                  if (d.x < p.x) then Direction := Left;
                  if (d.x > p.x) then Direction := Right;
                  if (d.y < p.y) then Direction := Up;
                  if (d.y > p.y) then Direction := Down;
                  case LastDirection of
                    Left:
                      begin
                        if (Self.Direction = Down) then LastTurn := Left;
                        if (Self.Direction = Up) then LastTurn := Right;
                      end;
                    Right:
                      begin
                        if (Self.Direction = Down) then LastTurn := Right;
                        if (Self.Direction = Up) then LastTurn := Left;
                      end;
                    Up:
                      begin
                        if (Self.Direction = Left) then LastTurn := Left;
                        if (Self.Direction = Right) then LastTurn := Right;
                      end;
                    Down:
                      begin
                        if (Self.Direction = Left) then LastTurn := Right;
                        if (Self.Direction = Right) then LastTurn := Left;
                      end;
                  end;
                  ALevel.SetSprite(d.x, d.y, Self);
                  ALevel.SetSprite(p.x, p.y, SpriteNothing.Create);
                  Break;
                end;
            end;
        end;

      {$IF NOT DEFINED(CLR)}
      for I := Low(w) to High(w) do
        FreeAndNil(w[I]);
      SetLength(w, 0);

      FreeAndNil(d);
      FreeAndNil(p);
      {$IFEND}

    end;

end;

{ Ghost90R }

procedure SpriteGhost90R.Move(ALevel: Level);
type
  TPositionArray = array of Position;
var
  w: TPositionArray;
  d, p: Position;
  I: Integer;
begin

{$IF NOT DEFINED(CLR)}
  d := nil; p := nil;
{$IFEND}

  if IsAlive then
    begin
      p := ALevel.GetSpritePosition(Self);
      SetLength(w, 4);
      for I := Low(w) to High(w) do
        w[I] := Position.Create(p);
      case Direction of
        Left:
          begin
            Dec(w[0].x);
            Inc(w[1].y);
            Dec(w[2].y);
            Inc(w[3].x);
          end;
        Right:
          begin
            Inc(w[0].x);
            Dec(w[1].y);
            Inc(w[2].y);
            Dec(w[3].x);
          end;
        Up:
          begin
            Dec(w[0].y);
            Dec(w[1].x);
            Inc(w[2].x);
            Inc(w[3].y);
          end;
        Down:
          begin
            Inc(w[0].y);
            Inc(w[1].x);
            Dec(w[2].x);
            Dec(w[3].y);
          end;
      end;
      for I := 0 to 3 do
        begin
          if not p.Equal(w[i]) then
            begin
              {$IF NOT DEFINED(CLR)}FreeAndNil(d);{$IFEND}
              d := Position.Create(w[i]);
              // Digger
              if ALevel.GetSprite(d.x, d.y) is SpriteDigger then
                ALevel.ADigger.Die;
              // Nothing
              if ALevel.GetSprite(d.x, d.y) is SpriteNothing then
                begin
                  if (d.x < p.x) then Direction := Left;
                  if (d.x > p.x) then Direction := Right;
                  if (d.y < p.y) then Direction := Up;
                  if (d.y > p.y) then Direction := Down;
                  ALevel.SetSprite(d.x, d.y, Self);
                  ALevel.SetSprite(p.x, p.y, SpriteNothing.Create);
                  Break;
                end;
            end;
        end;

      {$IF NOT DEFINED(CLR)}
      for I := Low(w) to High(w) do
        FreeAndNil(w[I]);
      SetLength(w, 0);

      FreeAndNil(d);
      FreeAndNil(p);
      {$IFEND}

    end;

end;

{ Wall }

function SpriteWall.get_ImageIndex: Integer;
begin
  Result := 14;
end;

procedure TDiggerPanel.Reset;
begin
  FLevel.Time := Time;
  FLevel.ResetLevel;
  FLevel.Load(FTime);
  Start;
end;

procedure TDiggerPanel.ResetLives;
begin
  FLives := FInitLives;
end;

procedure TDiggerPanel.Restart;
begin
  if FLives > 0 then
    begin
      Dec(FLives);
      FLevel.Load(FTime);
    end;
end;

procedure TDiggerPanel.SetLives(const Value: Integer);
begin
  if (FLives<>Value) and (Value<1000) then
    begin
      FLives := Value;
      FInitLives := Value;
      inherited Invalidate;
    end;
end;

procedure TDiggerPanel.SetMagnification(const Value: Integer);
begin
  if (Value <> FMagnification) and (Value>0) then
    begin
      FMagnification := Value;
      GMagnification := Value;
      FChangingMagnification := True;
      UpdateMagnification;
      FChangingMagnification := False;
    end;
end;

procedure TDiggerPanel.SetTime(const Value: Integer);
begin
  if FTime<>Value then
    begin
      if Assigned(FLevel) then
        FLevel.Time := Value;
      FTime := Value;
      inherited Invalidate;
    end;
end;

procedure TDiggerPanel.UpdateHookKeys;
var
{$IF NOT DEFINED(CLR)}
  AForm: TForm;
  MyKeyEvent, KeyEvent: TKeyEvent;
{$ELSE}
  AForm: System.Windows.Forms.Form;
{$IFEND}
begin
{$IF NOT DEFINED(CLR)}
  if Parent is TForm then
    AForm := Parent as TForm
{$ELSE}
  if Parent is System.Windows.Forms.Form then
    AForm := Parent as System.Windows.Forms.Form
{$IFEND}
    else
    AForm := nil;
  if FHookKeys then
    begin
      if Assigned(AForm) then
        begin
{$IF NOT DEFINED(CLR)}
          AForm.OnKeyUp := FormKeyUp;
          AForm.OnKeyDown := FormKeyDown;
{$ELSE}
          Include(AForm.KeyUp, FormKeyUp);
          Include(AForm.KeyDown, FormKeyDown);
{$IFEND}
        end;
    end else
    begin
      if Assigned(AForm) then
        begin
{$IF NOT DEFINED(CLR)}
          MyKeyEvent := FormKeyUp;
          KeyEvent := AForm.OnKeyUp;
          if TMethod(KeyEvent).Code = TMethod(MyKeyEvent).Code then
            AForm.OnKeyUp := nil;
          MyKeyEvent := FormKeyDown;
          KeyEvent := AForm.OnKeyDown;
          if TMethod(KeyEvent).Code = TMethod(MyKeyEvent).Code then
            AForm.OnKeyDown := nil;
{$ELSE}
          Exclude(AForm.KeyUp, FormKeyUp);
          Exclude(AForm.KeyDown, FormKeyDown);
{$IFEND}
        end;
    end;
end;

procedure TDiggerPanel.UpdateMagnification;
var
  ARect: Rectangle;
  SavedHeight: Integer;
  AForm: {$IF DEFINED(CLR)}System.Windows.Forms.Form{$ELSE}TWinControl{$IFEND};
begin
  AForm := {$IF DEFINED(CLR)}FindForm{$ELSE}Parent{$IFEND};
  if Assigned(AForm) then
{$IF NOT DEFINED(CLR)}
    SavedHeight := AForm.ClientHeight - Height else
{$ELSE}
    SavedHeight := AForm.ClientSize.Height - Height else
{$IFEND}
    SavedHeight := 0;
  ARect := MagnifyRectangle(0, 0, 320, 256);
  Width  := ARect.Right;
  Height := ARect.Bottom;
  InitializeSettings;
  if Assigned(FOnMagnificationChanged) then
    FOnMagnificationChanged(Self) else
  if Assigned(AForm) then
    begin
{$IF NOT DEFINED(CLR)}
      OutputDebugString('Changing Form Size');
      AForm.ClientWidth  := Width;
      AForm.ClientHeight := Height + SavedHeight;
{$ELSE}
      AForm.ClientSize := System.Drawing.Size.Create(Width, Height);
      inherited Invalidate;
{$IFEND}
    end;
end;

procedure TDiggerPanel.SetHookKeys(const Value: Boolean);
begin
  if FHookKeys<>Value then
    FHookKeys := Value;
  if Assigned(Parent) then
    UpdateHookKeys;
end;


procedure TDiggerPanel.Pause;
begin
  Timer.Enabled := not Timer.Enabled;
end;

procedure TDiggerPanel.Start;
begin
  FGameStarted := True;
  Timer.Enabled := False;
  // FLevel.Number := 0;
  FLevel.Load(FTime);
  ResetLives;
  Timer.Enabled := True;
end;

{$IF NOT DEFINED(CLR)}
procedure TDiggerPanel.Timer_Tick(Sender: TObject);
{$ELSE}
procedure TDiggerPanel.Timer_Tick(sender: &Object; pe: EventArgs);
{$IFEND}
var
  I, Y, X: Integer;
  LSprite: Sprite;
  LRectangle: Rectangle;
begin
  Inc(tick);
  if not Assigned(FLevel) then
    begin
      {$IF NOT DEFINED(CLR)}asm nop end;{$IFEND}
      exit;
    end;
  FLevel.Tick1;
  if ((tick mod 2) = 0) then
    begin
      for I := Low(keysRelease) to High(keysRelease) do
        if keysRelease[I] then
          begin
            FLevel.ADigger.Keys[I] := False;
            keysRelease[I] := False;
          end;
      FLevel.Tick4;
    end;
  Invalidate(MagnifyRectangle(0, 8, 320, 161));
  for Y := 0 to 13 do
    begin
      for X := 0 to 19 do
        begin
          LSprite := FLevel.GetSprite(x, y);
          if LSprite.Invalidate then
            begin
              LRectangle := MagnifyRectangle(x * 16, y * 16 + 32, 16, 16);
              Invalidate(LRectangle);
              LSprite.Invalidate := False;
            end;
        end;
    end;
  Update;
end;

{$IF DEFINED(MSWINDOWS)}
procedure TDiggerPanel.WMEraseBkgnd(var Message: TMessage);
begin
  Message.Result := 1;
end;
{$IFEND}

end.

